# Copyright (c) 2018-2021 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
import logging
from typing import Dict, Optional, NamedTuple

LOGGER = logging.getLogger("transmission-rpc")
LOGGER.setLevel(logging.ERROR)


def mirror_dict(source: dict) -> dict:
    """
    Creates a dictionary with all values as keys and all keys as values.
    """
    source.update({value: key for key, value in source.items()})
    return source


DEFAULT_TIMEOUT = 30.0

TR_PRI_LOW = -1
TR_PRI_NORMAL = 0
TR_PRI_HIGH = 1

PRIORITY = mirror_dict({"low": TR_PRI_LOW, "normal": TR_PRI_NORMAL, "high": TR_PRI_HIGH})

TR_RATIOLIMIT_GLOBAL = 0  # follow the global settings
TR_RATIOLIMIT_SINGLE = 1  # override the global settings, seeding until a certain ratio
TR_RATIOLIMIT_UNLIMITED = 2  # override the global settings, seeding regardless of ratio

RATIO_LIMIT = mirror_dict(
    {
        "global": TR_RATIOLIMIT_GLOBAL,
        "single": TR_RATIOLIMIT_SINGLE,
        "unlimited": TR_RATIOLIMIT_UNLIMITED,
    }
)

TR_IDLELIMIT_GLOBAL = 0  # follow the global settings
TR_IDLELIMIT_SINGLE = 1  # override the global settings, seeding until a certain idle time
TR_IDLELIMIT_UNLIMITED = 2  # override the global settings, seeding regardless of activity

IDLE_LIMIT = mirror_dict(
    {
        "global": TR_RATIOLIMIT_GLOBAL,
        "single": TR_RATIOLIMIT_SINGLE,
        "unlimited": TR_RATIOLIMIT_UNLIMITED,
    }
)


class Args(NamedTuple):
    type: str
    added_version: int
    removed_version: Optional[int] = None
    previous_argument_name: Optional[str] = None
    next_argument_name: Optional[str] = None
    description: str = ""

    def __repr__(self) -> str:
        return (
            f"Args({self.type!r},"
            f" {self.added_version!r},"
            f" {self.removed_version!r},"
            f" {self.previous_argument_name!r},"
            f" {self.next_argument_name!r},"
            f" {self.description!r})"
        )

    def __str__(self) -> str:
        return f"Args<type={self.type}, {self.added_version}, description={self.description!r})"


class Type:
    number = "number"
    string = "string"
    array = "array"
    boolean = "boolean"
    double = "double"
    object = "object"


TORRENT_GET_ARGS: Dict[str, Args] = {
    "activityDate": Args(Type.number, 1, description="Last time of upload or download activity."),
    "addedDate": Args(Type.number, 1, description="The date when this torrent was first added."),
    "announceResponse": Args(Type.string, 1, 7, description="The announce message from the tracker."),
    "announceURL": Args(Type.string, 1, 7, description="Current announce URL."),
    "comment": Args(Type.string, 1, description="Torrent comment."),
    "corruptEver": Args(Type.number, 1, description="Number of bytes of corrupt data downloaded."),
    "creator": Args(Type.string, 1, description="Torrent creator."),
    "dateCreated": Args(Type.number, 1, description="Torrent creation date."),
    "desiredAvailable": Args(Type.number, 1, description="Number of bytes available and left to be downloaded."),
    "doneDate": Args(Type.number, 1, description="The date when the torrent finished downloading."),
    "downloadedEver": Args(Type.number, 1, description="Number of bytes of good data downloaded."),
    "downloadLimit": Args(Type.number, 1, None, None, None, "Download limit in Kbps."),
    "downloadLimitMode": Args(
        Type.number, 1, 5, description="Download limit mode. 0 means global, 1 means single, 2 unlimited."
    ),
    "error": Args(
        Type.number,
        1,
        description="Kind of error. 0 means OK, 1 means tracker warning, 2 means tracker error, 3 means local error.",
    ),
    "errorString": Args(Type.number, 1, None, None, None, "Error message."),
    "eta": Args(
        Type.number,
        1,
        description="Estimated number of seconds left when downloading or seeding. -1 means not available and -2 means unknown.",
    ),
    "files": Args(Type.array, 1, description="Array of file object containing key, bytesCompleted, length and name."),
    "hashString": Args(Type.string, 1, description="Hashstring unique for the torrent even between sessions."),
    "haveUnchecked": Args(Type.number, 1, None, None, None, "Number of bytes of partial pieces."),
    "haveValid": Args(Type.number, 1, description="Number of bytes of checksum verified data."),
    "id": Args(Type.number, 1, None, None, None, "Session unique torrent id."),
    "isPrivate": Args(Type.boolean, 1, None, None, None, "True if the torrent is private."),
    "lastAnnounceTime": Args(Type.number, 1, 7, None, None, "The time of the last announcement."),
    "lastScrapeTime": Args(Type.number, 1, 7, None, None, "The time af the last successful scrape."),
    "leechers": Args(Type.number, 1, 7, None, None, "Number of leechers."),
    "leftUntilDone": Args(Type.number, 1, description="Number of bytes left until the download is done."),
    "manualAnnounceTime": Args(Type.number, 1, description="The time until you manually ask for more peers."),
    "maxConnectedPeers": Args(Type.number, 1, None, None, None, "Maximum of connected peers."),
    "name": Args(Type.string, 1, None, None, None, "Torrent name."),
    "nextAnnounceTime": Args(Type.number, 1, 7, None, None, "Next announce time."),
    "nextScrapeTime": Args(Type.number, 1, 7, None, None, "Next scrape time."),
    "peersConnected": Args(Type.number, 1, None, None, None, "Number of peers we are connected to."),
    "peersFrom": Args(Type.object, 1, description="Object containing download peers counts for different peer types."),
    "peersGettingFromUs": Args(Type.number, 1, description="Number of peers we are sending data to."),
    "peersKnown": Args(Type.number, 1, 13, description="Number of peers that the tracker knows."),
    "peersSendingToUs": Args(Type.number, 1, None, None, None, "Number of peers sending to us"),
    "percentComplete": Args(Type.double, 17),
    "pieceCount": Args(Type.number, 1, None, None, None, "Number of pieces."),
    "pieceSize": Args(Type.number, 1, None, None, None, "Number of bytes in a piece."),
    "priorities": Args(Type.array, 1, None, None, None, "Array of file priorities."),
    "rateDownload": Args(Type.number, 1, None, None, None, "Download rate in bps."),
    "rateUpload": Args(Type.number, 1, None, None, None, "Upload rate in bps."),
    "recheckProgress": Args(Type.double, 1, None, None, None, "Progress of recheck. 0.0 to 1.0."),
    "scrapeResponse": Args(Type.string, 1, 7, None, None, "Scrape response message."),
    "scrapeURL": Args(Type.string, 1, 7, None, None, "Current scrape URL"),
    "seeders": Args(Type.number, 1, 7, description="Number of seeders reported by the tracker."),
    "sizeWhenDone": Args(Type.number, 1, description="Size of the torrent download in bytes."),
    "startDate": Args(Type.number, 1, description="The date when the torrent was last started."),
    "status": Args(Type.number, 1, None, None, None, "Current status, see source"),
    "swarmSpeed": Args(Type.number, 1, 7, None, None, "Estimated speed in Kbps in the swarm."),
    "timesCompleted": Args(Type.number, 1, 7, description="Number of successful downloads reported by the tracker."),
    "trackers": Args(Type.array, 1, None, None, None, "Array of tracker objects."),
    "totalSize": Args(Type.number, 1, None, None, None, "Total size of the torrent in bytes"),
    "uploadedEver": Args(Type.number, 1, None, None, None, "Number of bytes uploaded, ever."),
    "uploadLimit": Args(Type.number, 1, None, None, None, "Upload limit in Kbps"),
    "uploadLimitMode": Args(
        Type.number, 1, 5, description="Upload limit mode. 0 means global, 1 means single, 2 unlimited."
    ),
    "uploadRatio": Args(Type.double, 1, None, None, None, "Seed ratio."),
    "wanted": Args(Type.array, 1, description="Array of booleans indicated wanted files."),
    "webseeds": Args(Type.array, 1, None, None, None, "Array of webseeds objects"),
    "webseedsSendingToUs": Args(Type.number, 1, None, None, None, "Number of webseeds seeding to us."),
    "peers": Args(Type.array, 2, None, None, None, "Array of peer objects."),
    "downloadDir": Args(Type.string, 4, description="The directory path where the torrent is downloaded to."),
    "downloaders": Args(Type.number, 4, 7, None, None, "Number of downloaders."),
    "bandwidthPriority": Args(Type.number, 5, description="Bandwidth priority. Low (-1), Normal (0) or High (1)."),
    "downloadLimited": Args(Type.boolean, 5, None, None, None, "Download limit is enabled"),
    "fileStats": Args(
        Type.array, 5, description="Away of file statistics containing bytesCompleted, wanted and priority."
    ),
    "honorsSessionLimits": Args(Type.boolean, 5, description="True if session upload limits are honored"),
    "peer-limit": Args(Type.number, 5, None, None, None, "Maximum number of peers."),
    "percentDone": Args(Type.double, 5, description="Download progress of selected files. 0.0 to 1.0."),
    "pieces": Args(Type.string, 5, description="String with base64 encoded bitfield indicating finished pieces."),
    "seedRatioLimit": Args(Type.double, 5, None, None, None, "Seed ratio limit."),
    "seedRatioMode": Args(Type.number, 5, description="Use global (0), torrent (1), or unlimited (2) limit."),
    "torrentFile": Args(Type.string, 5, None, None, None, "Path to .torrent file."),
    "uploadLimited": Args(Type.boolean, 5, None, None, None, "Upload limit enabled."),
    "magnetLink": Args(Type.string, 7, None, None, None, "The magnet link for this torrent."),
    "metadataPercentComplete": Args(Type.number, 7, description="Download progress of metadata. 0.0 to 1.0."),
    "trackerStats": Args(Type.object, 7, description="Array of object containing tracker statistics."),
    "isFinished": Args(Type.boolean, 9, description="True if the torrent is finished. Downloaded and seeded."),
    "seedIdleLimit": Args(Type.number, 10, None, None, None, "Idle limit in minutes."),
    "seedIdleMode": Args(Type.number, 10, description="Use global (0), torrent (1), or unlimited (2) limit."),
    "isStalled": Args(Type.boolean, 14, description="True if the torrent has stalled (been idle for a long time)."),
    "queuePosition": Args(Type.number, 14, None, None, None, "The queue position."),
    "etaIdle": Args(
        Type.number,
        15,
        description="Estimated number of seconds left until the idle time limit is reached. -1 means not available and -2 means unknown.",
    ),
    "secondsDownloading": Args(Type.number, 15, None, None, None, ""),
    "secondsSeeding": Args(Type.number, 15, None, None, None, ""),
    "labels": Args(Type.array, 16, None, None, None, "array of string labels"),
    "editDate": Args(Type.number, 16),
    "file-count": Args(Type.number, 17),
    "primary-mime-type": Args(Type.string, 17),
    "group": Args(Type.string, 17, description="The name of this torrent's bandwidth group"),
    "trackerList": Args(
        Type.array,
        17,
        description="string of announce URLs, one per line, with a blank line between tiers",
    ),
}

TORRENT_SET_ARGS: Dict[str, Args] = {
    "files-wanted": Args(Type.array, 1, description="A list of file id's that should be downloaded."),
    "files-unwanted": Args(Type.array, 1, description="A list of file id's that shouldn't be downloaded."),
    "location": Args(Type.array, 1, None, None, None, "Local download location."),
    "peer-limit": Args(Type.number, 1, None, None, None, "The peer limit for the torrents."),
    "priority-high": Args(Type.array, 1, description="A list of file id's that should have high priority."),
    "priority-low": Args(Type.array, 1, description="A list of file id's that should have normal priority."),
    "priority-normal": Args(Type.array, 1, description="A list of file id's that should have low priority."),
    "speed-limit-down": Args(Type.number, 1, 5, None, "downloadLimit", "Set the speed limit for download in Kib/s."),
    "speed-limit-down-enabled": Args(Type.boolean, 1, 5, None, "downloadLimited", "Enable download speed limiter."),
    "speed-limit-up": Args(Type.number, 1, 5, None, "uploadLimit", "Set the speed limit for upload in Kib/s."),
    "speed-limit-up-enabled": Args(Type.boolean, 1, 5, None, "uploadLimited", "Enable upload speed limiter."),
    "bandwidthPriority": Args(Type.number, 5, None, None, None, "Priority for this transfer."),
    "downloadLimit": Args(Type.number, 5, None, "speed-limit-down", None, "Set the speed limit for download in Kib/s."),
    "downloadLimited": Args(Type.boolean, 5, None, "speed-limit-down-enabled", None, "Enable download speed limiter."),
    "honorsSessionLimits": Args(
        Type.boolean,
        5,
        description="Enables or disables the transfer to honour the upload limit set in the session.",
    ),
    "seedRatioLimit": Args(Type.double, 5, None, None, None, "Seeding ratio."),
    "seedRatioMode": Args(
        Type.number,
        5,
        description="Which ratio to use. 0 = Use session limit, 1 = Use transfer limit, 2 = Disable limit.",
    ),
    "uploadLimit": Args(Type.number, 5, None, "speed-limit-up", None, "Set the speed limit for upload in Kib/s."),
    "uploadLimited": Args(Type.boolean, 5, None, "speed-limit-up-enabled", None, "Enable upload speed limiter."),
    "seedIdleLimit": Args(Type.number, 10, None, None, None, "Seed inactivity limit in minutes."),
    "seedIdleMode": Args(
        Type.number,
        10,
        description="Seed inactivity mode. 0 = Use session limit, 1 = Use transfer limit, 2 = Disable limit.",
    ),
    "trackerAdd": Args(Type.array, 10, removed_version=17, description="Array of string with announce URLs to add."),
    "trackerRemove": Args(Type.array, 10, removed_version=17, description="Array of ids of trackers to remove."),
    "trackerReplace": Args(
        Type.array, 10, 17, description="Array of (id, url) tuples where the announce URL should be replaced."
    ),
    "queuePosition": Args(Type.number, 14, description="Position of this transfer in its queue."),
    "labels": Args(Type.array, 16, None, None, None, "array of string labels"),
    "trackerList": Args(
        Type.array, 17, description="string of announce URLs, one per line, with a blank line between tiers"
    ),
}
TORRENT_ADD_ARGS: Dict[str, Args] = {
    "download-dir": Args(Type.string, 1, description="The directory where the downloaded contents will be saved in."),
    "filename": Args(Type.string, 1, description="A file path or URL to a torrent file or a magnet link."),
    "files-wanted": Args(Type.array, 1, description="A list of file id's that should be downloaded."),
    "files-unwanted": Args(Type.array, 1, description="A list of file id's that shouldn't be downloaded."),
    "metainfo": Args(Type.string, 1, description="The content of a torrent file, base64 encoded."),
    "paused": Args(Type.boolean, 1, description="If True, does not start the transfer when added."),
    "peer-limit": Args(Type.number, 1, None, None, None, "Maximum number of peers allowed."),
    "priority-high": Args(Type.array, 1, description="A list of file id's that should have high priority."),
    "priority-low": Args(Type.array, 1, description="A list of file id's that should have low priority."),
    "priority-normal": Args(Type.array, 1, description="A list of file id's that should have normal priority."),
    "bandwidthPriority": Args(Type.number, 8, None, None, None, "Priority for this transfer."),
    "cookies": Args(Type.string, 13, None, None, None, "One or more HTTP cookie(s)."),
    "trackerList": Args(
        Type.array, 17, description="string of announce URLs, one per line, with a blank line between tiers"
    ),
}

SESSION_GET_ARGS = {
    "download-dir": Args(Type.string, 1, None, None, None, "The download directory."),
    "encryption": Args(
        Type.string, 1, description="Encryption mode, one of ``required``, ``preferred`` or ``tolerated``."
    ),
    "peer-limit": Args(Type.number, 1, 5, None, "peer-limit-global", "Maximum number of peers."),
    "pex-allowed": Args(Type.boolean, 1, 5, None, "pex-enabled", "True if PEX is allowed."),
    "port": Args(Type.number, 1, 5, None, "peer-port", "Peer port."),
    "port-forwarding-enabled": Args(Type.boolean, 1, None, None, None, "True if port forwarding is enabled."),
    "speed-limit-down": Args(Type.number, 1, None, None, None, "Download speed limit (in Kib/s)."),
    "speed-limit-down-enabled": Args(Type.boolean, 1, description="True if the download speed is limited."),
    "speed-limit-up": Args(Type.number, 1, None, None, None, "Upload speed limit (in Kib/s)."),
    "speed-limit-up-enabled": Args(Type.boolean, 1, description="True if the upload speed is limited."),
    "version": Args(Type.string, 3, None, None, None, "Transmission version."),
    "rpc-version": Args(Type.number, 4, None, None, None, "Transmission RPC API Version."),
    "rpc-version-minimum": Args(Type.number, 4, None, None, None, "Minimum accepted RPC API Version."),
    "alt-speed-down": Args(Type.number, 5, description="Alternate session download speed limit (in Kib/s)."),
    "alt-speed-enabled": Args(
        Type.boolean, 5, description="True if alternate global download speed limiter is enabled."
    ),
    "alt-speed-time-begin": Args(
        Type.number, 5, description="Time when alternate speeds should be enabled. Minutes after midnight."
    ),
    "alt-speed-time-enabled": Args(Type.boolean, 5, description="True if alternate speeds scheduling is enabled."),
    "alt-speed-time-end": Args(
        Type.number, 5, description="Time when alternate speeds should be disabled. Minutes after midnight."
    ),
    "alt-speed-time-day": Args(Type.number, 5, description="Days alternate speeds scheduling is enabled."),
    "alt-speed-up": Args(Type.number, 5, description="Alternate session upload speed limit (in Kib/s)"),
    "blocklist-enabled": Args(Type.boolean, 5, None, None, None, "True when blocklist is enabled."),
    "blocklist-size": Args(Type.number, 5, None, None, None, "Number of rules in the blocklist"),
    "peer-limit-global": Args(Type.number, 5, None, "peer-limit", None, "Maximum number of peers."),
    "peer-limit-per-torrent": Args(Type.number, 5, description="Maximum number of peers per transfer."),
    "pex-enabled": Args(Type.boolean, 5, None, "pex-allowed", None, "True if PEX is enabled."),
    "peer-port": Args(Type.number, 5, None, "port", None, "Peer port."),
    "peer-port-random-on-start": Args(
        Type.boolean, 5, description="Enables randomized peer port on start of Transmission."
    ),
    "seedRatioLimit": Args(Type.double, 5, description="Seed ratio limit. 1.0 means 1:1 download and upload ratio."),
    "seedRatioLimited": Args(Type.boolean, 5, description="True if seed ration limit is enabled."),
    "dht-enabled": Args(Type.boolean, 6, None, None, None, "True if DHT enabled."),
    "incomplete-dir": Args(
        Type.string, 7, description="The path to the directory for incomplete torrent transfer data."
    ),
    "incomplete-dir-enabled": Args(Type.boolean, 7, description="True if the incomplete dir is enabled."),
    "config-dir": Args(Type.string, 8, description="location of transmissions configuration directory"),
    "rename-partial-files": Args(Type.boolean, 8, description='True if ".part" is appended to incomplete files'),
    "lpd-enabled": Args(Type.boolean, 9, description="True if local peer discovery is enabled."),
    "script-torrent-done-enabled": Args(Type.boolean, 9, None, None, None, "True if the done script is enabled."),
    "script-torrent-done-filename": Args(
        Type.string, 9, description="Filename of the script to run when the transfer is done."
    ),
    "start-added-torrents": Args(Type.boolean, 9, description="When true uploaded torrents will start right away."),
    "trash-original-torrent-files": Args(
        Type.boolean, 9, description="When true added .torrent files will be deleted."
    ),
    "cache-size-mb": Args(Type.number, 10, description="The maximum size of the disk cache in MB"),
    "idle-seeding-limit": Args(Type.number, 10, None, None, None, "Seed inactivity limit in minutes."),
    "idle-seeding-limit-enabled": Args(Type.boolean, 10, description="True if the seed activity limit is enabled."),
    "units": Args(Type.object, 10, description="An object containing units for size and speed."),
    "blocklist-url": Args(Type.string, 11, description="Location of the block list. Updated with blocklist-update."),
    "download-dir-free-space": Args(
        Type.number, 12, 17, None, "free-spaces", "Free space in the download directory, in bytes"
    ),
    "utp-enabled": Args(Type.boolean, 13, description="True if Micro Transport Protocol (UTP) is enabled."),
    "download-queue-size": Args(Type.number, 14, description="Number of slots in the download queue."),
    "download-queue-enabled": Args(Type.boolean, 14, description="True if the download queue is enabled."),
    "queue-stalled-minutes": Args(
        Type.number, 14, description="Number of minutes of idle that marks a transfer as stalled."
    ),
    "queue-stalled-enabled": Args(Type.boolean, 14, description="True if stalled tracking of transfers is enabled."),
    "seed-queue-size": Args(Type.number, 14, description="Number of slots in the upload queue."),
    "seed-queue-enabled": Args(Type.boolean, 14, None, None, None, "True if upload queue is enabled."),
    "fields": Args(Type.array, 16, None, None, None, "array of keys"),
    "session-id": Args(Type.string, 16, None, None, None, "session-id"),
}

SESSION_SET_ARGS = {
    "download-dir": Args(Type.string, 1, None, None, None, "Set the session download directory."),
    "encryption": Args(
        Type.string,
        1,
        description="Set the session encryption mode, one of ``required``, ``preferred`` or ``tolerated``.",
    ),
    "peer-limit": Args(Type.number, 1, 5, None, "peer-limit-global", "Maximum number of peers."),
    "pex-allowed": Args(Type.boolean, 1, 5, None, "pex-enabled", "Allowing PEX in public torrents."),
    "port": Args(Type.number, 1, 5, None, "peer-port", "Peer port."),
    "port-forwarding-enabled": Args(Type.boolean, 1, None, None, None, "Enables port forwarding."),
    "speed-limit-down": Args(Type.number, 1, None, None, None, "Download speed limit (in Kib/s)."),
    "speed-limit-down-enabled": Args(Type.boolean, 1, None, None, None, "Enables download speed limiting."),
    "speed-limit-up": Args(Type.number, 1, None, None, None, "Upload speed limit (in Kib/s)."),
    "speed-limit-up-enabled": Args(Type.boolean, 1, None, None, None, "Enables upload speed limiting."),
    "alt-speed-down": Args(Type.number, 5, description="Alternate session download speed limit (in Kib/s)."),
    "alt-speed-enabled": Args(Type.boolean, 5, description="Enables alternate global download speed limiter."),
    "alt-speed-time-begin": Args(
        Type.number, 5, description="Time when alternate speeds should be enabled. Minutes after midnight."
    ),
    "alt-speed-time-enabled": Args(Type.boolean, 5, description="Enables alternate speeds scheduling."),
    "alt-speed-time-end": Args(
        Type.number, 5, description="Time when alternate speeds should be disabled. Minutes after midnight."
    ),
    "alt-speed-time-day": Args(Type.number, 5, description="Enables alternate speeds scheduling these days."),
    "alt-speed-up": Args(Type.number, 5, description="Alternate session upload speed limit (in Kib/s)."),
    "blocklist-enabled": Args(Type.boolean, 5, None, None, None, "Enables the block list"),
    "peer-limit-global": Args(Type.number, 5, None, "peer-limit", None, "Maximum number of peers."),
    "peer-limit-per-torrent": Args(Type.number, 5, description="Maximum number of peers per transfer."),
    "pex-enabled": Args(Type.boolean, 5, None, "pex-allowed", None, "Allowing PEX in public torrents."),
    "peer-port": Args(Type.number, 5, None, "port", None, "Peer port."),
    "peer-port-random-on-start": Args(
        Type.boolean, 5, description="Enables randomized peer port on start of Transmission."
    ),
    "seedRatioLimit": Args(Type.double, 5, description="Seed ratio limit. 1.0 means 1:1 download and upload ratio."),
    "seedRatioLimited": Args(Type.boolean, 5, None, None, None, "Enables seed ration limit."),
    "dht-enabled": Args(Type.boolean, 6, None, None, None, "Enables DHT."),
    "incomplete-dir": Args(Type.string, 7, description="The path to the directory of incomplete transfer data."),
    "incomplete-dir-enabled": Args(
        Type.boolean,
        7,
        description="Enables the incomplete transfer data directory. Otherwise data for incomplete transfers are stored in the download target.",
    ),
    "rename-partial-files": Args(Type.boolean, 8, None, None, None, 'Appends ".part" to incomplete files'),
    "lpd-enabled": Args(Type.boolean, 9, description="Enables local peer discovery for public torrents."),
    "script-torrent-done-enabled": Args(Type.boolean, 9, description='Whether or not to call the "done" script.'),
    "script-torrent-done-filename": Args(
        Type.string, 9, description="Filename of the script to run when the transfer is done."
    ),
    "start-added-torrents": Args(Type.boolean, 9, description="Added torrents will be started right away."),
    "trash-original-torrent-files": Args(
        Type.boolean, 9, description="The .torrent file of added torrents will be deleted."
    ),
    "cache-size-mb": Args(Type.number, 10, description="The maximum size of the disk cache in MB"),
    "idle-seeding-limit": Args(Type.number, 10, description="The default seed inactivity limit in minutes."),
    "idle-seeding-limit-enabled": Args(Type.boolean, 10, description="Enables the default seed inactivity limit"),
    "blocklist-url": Args(Type.string, 11, description="Location of the block list. Updated with blocklist-update."),
    "utp-enabled": Args(Type.boolean, 13, description="Enables Micro Transport Protocol (UTP)."),
    "download-queue-size": Args(Type.number, 14, description="Number of slots in the download queue."),
    "download-queue-enabled": Args(Type.boolean, 14, None, None, None, "Enables download queue."),
    "queue-stalled-minutes": Args(
        Type.number, 14, description="Number of minutes of idle that marks a transfer as stalled."
    ),
    "queue-stalled-enabled": Args(Type.boolean, 14, description="Enable tracking of stalled transfers."),
    "seed-queue-size": Args(Type.number, 14, description="Number of slots in the upload queue."),
    "seed-queue-enabled": Args(Type.boolean, 14, None, None, None, "Enables upload queue."),
}

# Arguments for torrent methods
TORRENT_ARGS: Dict[str, Dict[str, Args]] = {
    "get": TORRENT_GET_ARGS,
    "set": TORRENT_SET_ARGS,
    "add": TORRENT_ADD_ARGS,
}

# Arguments for session methods
SESSION_ARGS: Dict[str, Dict[str, Args]] = {
    "get": SESSION_GET_ARGS,
    "set": SESSION_SET_ARGS,
}



# Copyright (c) 2018-2021 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
import base64
import pathlib
import datetime
import warnings
from typing import Any, Dict, List, Tuple, Union, TypeVar, BinaryIO, Callable, Optional
from urllib.parse import urlparse

from transmission_rpc import constants
from transmission_rpc.error import TransmissionVersionError
from transmission_rpc.constants import LOGGER, Type

UNITS = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"]


def format_size(size: int) -> Tuple[float, str]:
    """
    Format byte size into IEC prefixes, B, KiB, MiB ...
    """
    s = float(size)
    i = 0
    while s >= 1024.0 and i < len(UNITS):
        i += 1
        s /= 1024.0
    return s, UNITS[i]


def format_speed(size: int) -> Tuple[float, str]:
    """
    Format bytes per second speed into IEC prefixes, B/s, KiB/s, MiB/s ...
    """
    (s, unit) = format_size(size)
    return s, unit + "/s"


def format_timedelta(delta: datetime.timedelta) -> str:
    """
    Format datetime.timedelta into <days> <hours>:<minutes>:<seconds>.
    """
    minutes, seconds = divmod(delta.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return f"{delta.days:d} {hours:02d}:{minutes:02d}:{seconds:02d}"


def rpc_bool(arg: Any) -> int:
    """
    Convert between Python boolean and Transmission RPC boolean.
    """
    if isinstance(arg, str):
        try:
            arg = bool(int(arg))
        except ValueError:
            arg = arg.lower() in ["true", "yes"]
    return 1 if bool(arg) else 0


TR_TYPE_MAP: Dict[str, Callable] = {
    Type.number: int,
    Type.string: str,
    Type.double: float,
    Type.boolean: rpc_bool,
    Type.array: list,
    Type.object: dict,
}


def make_python_name(name: str) -> str:
    """
    Convert Transmission RPC name to python compatible name.
    """
    return name.replace("-", "_")


def make_rpc_name(name: str) -> str:
    """
    Convert python compatible name to Transmission RPC name.
    """
    return name.replace("_", "-")


def argument_value_convert(
    method: str,
    argument: str,
    value: Any,
    rpc_version: int,
) -> Tuple[str, Any]:
    """
    Check and fix Transmission RPC issues with regards to methods, arguments and values.
    """
    if method in ("torrent-add", "torrent-get", "torrent-set"):
        args = constants.TORRENT_ARGS[method[-3:]]
    elif method in ("session-get", "session-set"):
        args = constants.SESSION_ARGS[method[-3:]]
    else:
        raise ValueError(f'Method "{method}" not supported')
    if argument in args:
        info = args[argument]
        invalid_version = True
        while invalid_version:
            invalid_version = False
            replacement = None
            if rpc_version < info.added_version:
                invalid_version = True
                replacement = info.previous_argument_name
            if info.removed_version is not None and info.removed_version <= rpc_version:
                invalid_version = True
                replacement = info.next_argument_name
            if invalid_version:
                if replacement:
                    LOGGER.warning(
                        'Replacing requested argument "%s" with "%s".',
                        argument,
                        replacement,
                    )
                    argument = replacement
                    info = args[argument]
                else:
                    raise ValueError(
                        f'Method "{method}" Argument "{argument}" does not exist in version {rpc_version:d}.'
                    )
        return argument, TR_TYPE_MAP[info.type](value)
    raise ValueError(f'Argument "{argument}" does not exists for method "{method}".')


def get_arguments(method: str, rpc_version: int) -> List[str]:
    """
    Get arguments for method in specified Transmission RPC version.
    """
    if method in ("torrent-add", "torrent-get", "torrent-set"):
        args = constants.TORRENT_ARGS[method[-3:]]
    elif method in ("session-get", "session-set"):
        args = constants.SESSION_ARGS[method[-3:]]
    else:
        raise ValueError(f'Method "{method}" not supported')
    accessible = []
    for argument, info in args.items():
        valid_version = True
        if rpc_version < info[1]:
            valid_version = False
        if info[2] is not None and info[2] <= rpc_version:
            valid_version = False
        if valid_version:
            accessible.append(argument)
    return accessible


_Fn = TypeVar("_Fn")


def _rpc_version_check(method: str, kwargs: Dict[str, Any], rpc_version: int) -> None:
    if method in ("torrent-add", "torrent-get", "torrent-set"):
        rpc_args = constants.TORRENT_ARGS[method[-3:]]
    elif method in ("session-get", "session-set"):
        rpc_args = constants.SESSION_ARGS[method[-3:]]
    else:
        raise ValueError(f'Method "{method}" not supported')

    for key, arg in rpc_args.items():
        if key in kwargs and arg.added_version > rpc_version:
            raise TransmissionVersionError(
                f'Method "{method}" Argument "{key}" does not exist in version {rpc_version}'
            )


def _try_read_torrent(torrent: Union[BinaryIO, str, bytes, pathlib.Path]) -> Optional[str]:  # pylint: disable=R0911
    """
    if torrent should be encoded with base64, return a non-None value.
    """
    # torrent is a str, may be a url
    if isinstance(torrent, str):
        parsed_uri = urlparse(torrent)
        # torrent starts with file, read from local disk and encode it to base64 url.
        if parsed_uri.scheme in ["https", "http", "magnet"]:
            return None

        if parsed_uri.scheme in ["file"]:
            warnings.warn("support for `file://` URL is deprecated.", DeprecationWarning)
            filepath = torrent
            # uri decoded different on linux / windows ?
            if len(parsed_uri.path) > 0:
                filepath = parsed_uri.path
            elif len(parsed_uri.netloc) > 0:
                filepath = parsed_uri.netloc
            with open(filepath, "rb") as torrent_file:
                return base64.b64encode(torrent_file.read()).decode("utf-8")

        # maybe it's base64 encoded file content
        try:
            # check if this is base64 data
            base64.b64decode(torrent.encode("utf-8"), validate=True)
            warnings.warn(
                "support for base64 encoded torrent content str is deprecated.",
                DeprecationWarning,
            )
            return torrent
        except (TypeError, ValueError):
            pass
    elif isinstance(torrent, pathlib.Path):
        return base64.b64encode(torrent.read_bytes()).decode("utf-8")
    elif isinstance(torrent, bytes):
        return base64.b64encode(torrent).decode("utf-8")
    # maybe a file, try read content and encode it.
    elif hasattr(torrent, "read"):
        return base64.b64encode(torrent.read()).decode("utf-8")

    return None



# Copyright (c) 2018-2021 Trim21 <i@trim21.me>
# Licensed under the MIT license.
from typing import Any, Tuple, Union, Optional, NamedTuple

from typing_extensions import Literal

_Number = Union[int, float]
_Timeout = Optional[Union[_Number, Tuple[_Number, _Number]]]


class File(NamedTuple):
    name: str  # file name
    size: int  # file size in bytes
    completed: int  # bytes completed
    priority: Literal["high", "normal", "low"]
    selected: bool  # if selected for download


class Field(NamedTuple):
    value: Any
    dirty: bool



"""
exception raise by this package
"""
# Copyright (c) 2018-2021 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
from typing import Optional

import requests.exceptions
from requests.models import Response


class TransmissionError(Exception):
    """
    This exception is raised when there has occurred an error related to
    communication with Transmission.
    """

    def __init__(self, message: str = "", original: Optional[Response] = None):
        super().__init__()
        self.message = message
        self.original = original

    def __str__(self) -> str:
        if self.original:
            original_name = type(self.original).__name__
            return f'{self.message} Original exception: {original_name}, "{self.original}"'
        return self.message


class TransmissionAuthError(TransmissionError):
    """Raised when username or password is incorrect"""


class TransmissionConnectError(TransmissionError, requests.exceptions.ConnectionError):
    """raised when client can't connect to transmission daemon"""


class TransmissionTimeoutError(TransmissionConnectError, requests.exceptions.Timeout):
    """Timeout"""


class TransmissionVersionError(TransmissionError):
    """transmission version is too lower to support some feature"""


class RemovedInV4Warning(PendingDeprecationWarning):
    """this method will be removed in transmission v4"""



# Copyright (c) 2018-2021 Trim21 <i@trim21.me>
# Copyright (c) 2020 littleya <me@littleya.com>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
import os
import re
import json
import time
import types
import string
import logging
import pathlib
import operator
import warnings
import urllib.parse
from typing import Any, Dict, List, Type, Tuple, Union, BinaryIO, Iterable, Optional
from warnings import warn
from urllib.parse import quote, urljoin

import requests
import requests.auth
import requests.exceptions
from typing_extensions import Literal

from transmission_rpc.error import (
    TransmissionError,
    RemovedInV4Warning,
    TransmissionAuthError,
    TransmissionConnectError,
    TransmissionTimeoutError,
)
from transmission_rpc.types import Group
from transmission_rpc.utils import (
    LOGGER,
    rpc_bool,
    get_arguments,
    make_rpc_name,
    _try_read_torrent,
    _rpc_version_check,
    argument_value_convert,
)
from transmission_rpc.session import Session
from transmission_rpc.torrent import Torrent
from transmission_rpc.constants import DEFAULT_TIMEOUT
from transmission_rpc.lib_types import File, Field, _Timeout

valid_hash_char = string.digits + string.ascii_letters

_TorrentID = Union[int, str]
_TorrentIDs = Union[str, _TorrentID, List[_TorrentID], None]


def ensure_location_str(s: Union[str, pathlib.Path]) -> str:
    if isinstance(s, pathlib.Path):
        if s.is_absolute():
            return str(s)

        warnings.warn(
            "`pathlib.Path` will be convert to absolute path automatically, "
            "please using a str or absolute Path to avoid unexpected path normalization\n"
            "This warning will become error in v4",
            DeprecationWarning,
        )
        return str(s.absolute())

    return str(s)


def _parse_torrent_id(raw_torrent_id: Union[int, str, Field]) -> Union[int, str]:
    if isinstance(raw_torrent_id, int):
        if raw_torrent_id >= 0:
            return raw_torrent_id
    elif isinstance(raw_torrent_id, str):
        if len(raw_torrent_id) != 40 or (set(raw_torrent_id) - set(valid_hash_char)):
            raise ValueError(f"torrent ids {raw_torrent_id} is not valid torrent id")
        return raw_torrent_id
    elif isinstance(raw_torrent_id, Field):
        return _parse_torrent_id(raw_torrent_id.value)
    raise ValueError(f"{raw_torrent_id} is not valid torrent id")


def _parse_torrent_ids(args: Any) -> Union[str, List[Union[str, int]]]:
    if args is None:
        return []
    if isinstance(args, int):
        return [_parse_torrent_id(args)]
    if isinstance(args, str):
        if args == "recently-active":
            return args
        return [_parse_torrent_id(args)]
    if isinstance(args, (list, tuple)):
        return [_parse_torrent_id(item) for item in args]
    raise ValueError(f"Invalid torrent id {args}")


class Client:
    def __init__(
        self,
        *,
        protocol: Literal["http", "https"] = "http",
        username: str = None,
        password: str = None,
        host: str = "127.0.0.1",
        port: int = 9091,
        path: str = "/transmission/",
        timeout: Union[int, float] = DEFAULT_TIMEOUT,
        logger: logging.Logger = LOGGER,
    ):
        if isinstance(logger, logging.Logger):
            self.logger = logger
        else:
            raise TypeError(
                "logger must be instance of `logging.Logger`, default: logging.getLogger('transmission-rpc')"
            )
        self._query_timeout: _Timeout = timeout

        username = quote(username or "", safe="$-_.+!*'(),;&=", encoding="utf8") if username else ""
        password = ":" + quote(password or "", safe="$-_.+!*'(),;&=", encoding="utf8") if password else ""
        auth = f"{username}{password}@" if (username or password) else ""
        url = urllib.parse.urlunparse((protocol, f"{auth}{host}:{port}", urljoin(path, "rpc"), None, None, None))
        self.url = str(url)
        self._sequence = 0
        self.session: Session = Session(self)
        self.session_id = "0"
        self.server_version: Optional[Tuple[int, int, Optional[str]]] = None
        self.protocol_version: Optional[int] = None
        self._http_session = requests.Session()
        self._http_session.trust_env = False
        self.get_session()
        self.torrent_get_arguments = get_arguments("torrent-get", self.rpc_version)

    @property
    def timeout(self) -> _Timeout:
        """
        Get current timeout for HTTP queries.
        """
        return self._query_timeout

    @timeout.setter
    def timeout(self, value: _Timeout) -> None:
        """
        Set timeout for HTTP queries.
        """
        if isinstance(value, (tuple, list)):
            if len(value) != 2:
                raise ValueError("timeout tuple can only include 2 numbers elements")
            for v in value:
                if not isinstance(v, (float, int)):
                    raise ValueError("element of timeout tuple can only be int of float")
            self._query_timeout = (value[0], value[1])  # for type checker
        elif value is None:
            self._query_timeout = DEFAULT_TIMEOUT
        else:
            self._query_timeout = float(value)

    @timeout.deleter
    def timeout(self) -> None:
        """
        Reset the HTTP query timeout to the default.
        """
        self._query_timeout = DEFAULT_TIMEOUT

    @property
    def _http_header(self) -> Dict[str, str]:
        return {"x-transmission-session-id": self.session_id}

    def _http_query(self, query: dict, timeout: _Timeout = None) -> str:
        """
        Query Transmission through HTTP.
        """
        request_count = 0
        if timeout is None:
            timeout = self.timeout
        while True:
            if request_count >= 10:
                raise TransmissionError("too much request, try enable logger to see what happened")
            self.logger.debug(
                {
                    "url": self.url,
                    "headers": self._http_header,
                    "data": query,
                    "timeout": timeout,
                }
            )
            request_count += 1
            try:
                r = self._http_session.post(
                    self.url,
                    headers=self._http_header,
                    json=query,
                    timeout=timeout,
                )
            except requests.exceptions.Timeout as e:
                raise TransmissionTimeoutError("timeout when connection to transmission daemon") from e
            except requests.exceptions.ConnectionError as e:
                raise TransmissionConnectError(f"can't connect to transmission daemon: {str(e)}") from e

            self.session_id = r.headers.get("X-Transmission-Session-Id", "0")
            self.logger.debug(r.text)
            if r.status_code in {401, 403}:
                self.logger.debug(r.request.headers)
                raise TransmissionAuthError("transmission daemon require auth", original=r)
            if r.status_code != 409:
                return r.text

    def _request(
        self,
        method: str,
        arguments: Dict[str, Any] = None,
        ids: _TorrentIDs = None,
        require_ids: bool = False,
        timeout: _Timeout = None,
    ) -> dict:
        """
        Send json-rpc request to Transmission using http POST
        """
        if not isinstance(method, str):
            raise ValueError("request takes method as string")
        if arguments is None:
            arguments = {}
        if not isinstance(arguments, dict):
            raise ValueError("request takes arguments as dict")
        arguments = {key.replace("_", "-"): value for key, value in arguments.items()}
        ids = _parse_torrent_ids(ids)
        if len(ids) > 0:
            arguments["ids"] = ids
        elif require_ids:
            raise ValueError("request require ids")

        query = {"tag": self._sequence, "method": method, "arguments": arguments}

        self._sequence += 1
        start = time.time()
        http_data = self._http_query(query, timeout)
        elapsed = time.time() - start
        self.logger.info("http request took %.3f s", elapsed)

        try:
            data: dict = json.loads(http_data)
        except ValueError as error:
            self.logger.error("Error: %s", str(error))
            self.logger.error('Request: "%s"', query)
            self.logger.error('HTTP data: "%s"', http_data)
            raise ValueError from error

        self.logger.debug(json.dumps(data, indent=2))
        if "result" in data:
            if data["result"] != "success":
                raise TransmissionError(f'Query failed with result "{data["result"]}".')
        else:
            raise TransmissionError("Query failed without result.")

        results = {}
        if method == "torrent-get":
            for item in data["arguments"]["torrents"]:
                results[item["id"]] = Torrent(self, item)
                if self.protocol_version == 2 and "peers" not in item:
                    self.protocol_version = 1
        elif method == "torrent-add":
            item = None
            if "torrent-added" in data["arguments"]:
                item = data["arguments"]["torrent-added"]
            elif "torrent-duplicate" in data["arguments"]:
                item = data["arguments"]["torrent-duplicate"]
            if item:
                results[item["id"]] = Torrent(self, item)
            else:
                raise TransmissionError("Invalid torrent-add response.")
        elif method == "session-get":
            self._update_session(data["arguments"])
        elif method == "session-stats":
            # older versions of T has the return data in "session-stats"
            if "session-stats" in data["arguments"]:
                self._update_session(data["arguments"]["session-stats"])
            else:
                self._update_session(data["arguments"])
        elif method in (
            "port-test",
            "blocklist-update",
            "free-space",
            "torrent-rename-path",
        ):
            results = data["arguments"]
        else:
            return data

        return results

    def _update_session(self, data: Dict[str, Any]) -> None:
        """
        Update session data.
        """
        if self.session:
            self.session._update(data)  # pylint: disable=W0212
        else:
            self.session = Session(self, data)

    def _update_server_version(self) -> None:
        """Decode the Transmission version string, if available."""
        if self.server_version is None:
            version_major = 2
            version_minor = 40
            version_change_set: Optional[str] = None
            version_parser = re.compile(r"(\d).(\d+) \((.*)\)")
            if hasattr(self.session, "version"):
                match = version_parser.match(self.session.version)
                if match:
                    version_major = int(match.group(1))
                    version_minor = int(match.group(2))
                    version_change_set = str(match.group(3))
            self.server_version = (version_major, version_minor, version_change_set)

    @property
    def rpc_version(self) -> int:
        """
        Get the Transmission RPC version. Trying to deduct if the server don't have a version value.
        """
        if self.protocol_version is None:
            # Ugly fix for 2.20 - 2.22 reporting rpc-version 11, but having new arguments
            if self.server_version and (self.server_version[0] == 2 and self.server_version[1] in [20, 21, 22]):
                self.protocol_version = 12
            # Ugly fix for 2.12 reporting rpc-version 10, but having new arguments
            elif self.server_version and (self.server_version[0] == 2 and self.server_version[1] == 12):
                self.protocol_version = 11
            elif hasattr(self.session, "rpc_version"):
                self.protocol_version = self.session.rpc_version
            elif hasattr(self.session, "version"):
                self.protocol_version = 3
            else:
                self.protocol_version = 2
        if self.server_version and (self.server_version[0] <= 2 and self.server_version[1] < 30):
            warnings.warn(
                "support for transmission version lower than 2.30 (rpc version 13) will be removed in the future",
                PendingDeprecationWarning,
            )
        return self.protocol_version

    def _rpc_version_warning(self, required_version: int) -> None:
        """
        Add a warning to the log if the Transmission RPC version is lower then the provided version.
        """
        if self.rpc_version < required_version:
            self.logger.warning(
                "Using feature not supported by server. RPC version for server %d, feature introduced in %d.",
                self.rpc_version,
                required_version,
            )

    def add_torrent(
        self,
        torrent: Union[BinaryIO, str, bytes, pathlib.Path],
        timeout: _Timeout = None,
        *,
        download_dir: str = None,
        files_unwanted: List[int] = None,
        files_wanted: List[int] = None,
        paused: bool = None,
        peer_limit: int = None,
        priority_high: List[int] = None,
        priority_low: List[int] = None,
        priority_normal: List[int] = None,
        cookies: str = None,
        labels: Optional[Iterable[str]] = None,
        bandwidthPriority: int = None,
    ) -> Torrent:
        """
        Add torrent to transfers list. ``torrent`` can be:

        - ``http://``, ``https://`` or  ``magnet:`` URL
        - torrent file-like object in binary mode
        - bytes of torrent content
        - ``pathlib.Path`` for local torrent file, will be read and encoded as base64.
        - **deprecated** str of base64 encoded torrent file content
        - **deprecated** ``file://`` URL

        .. NOTE::

            url starts with ``file://`` will be load by this package instead of transmission daemon

        Additional arguments are:

        ===================== ===== =========== =============================================================
        Argument              RPC   Replaced by Description
        ===================== ===== =========== =============================================================
        ``bandwidthPriority`` 8 -               Priority for this transfer.
        ``labels``            17 -              Array of string labels.
        ``cookies``           13 -              One or more HTTP cookie(s).
        ``download_dir``      1 -               The directory where the downloaded contents will be saved in.
        ``files_unwanted``    1 -               A list of file id's that shouldn't be downloaded.
        ``files_wanted``      1 -               A list of file id's that should be downloaded.
        ``paused``            1 -               If True, does not start the transfer when added.
        ``peer_limit``        1 -               Maximum number of peers allowed.
        ``priority_high``     1 -               A list of file id's that should have high priority.
        ``priority_low``      1 -               A list of file id's that should have low priority.
        ``priority_normal``   1 -               A list of file id's that should have normal priority.
        ===================== ===== =========== =============================================================

        Returns a Torrent object with the fields.
        """
        if torrent is None:
            raise ValueError("add_torrent requires data or a URI.")

        kwargs: Dict[str, Any] = {}
        if download_dir is not None:
            kwargs["download-dir"] = download_dir

        if files_unwanted is not None:
            kwargs["files-unwanted"] = files_unwanted

        if files_wanted is not None:
            kwargs["files-wanted"] = files_wanted

        if paused is not None:
            kwargs["paused"] = paused

        if peer_limit is not None:
            kwargs["peer-limit"] = peer_limit

        if priority_high is not None:
            kwargs["priority-high"] = priority_high

        if priority_low is not None:
            kwargs["priority-low"] = priority_low

        if priority_normal is not None:
            kwargs["priority-normal"] = priority_normal

        if bandwidthPriority is not None:
            kwargs["bandwidthPriority"] = bandwidthPriority

        if cookies is not None:
            kwargs["cookies"] = cookies

        if labels is not None:
            self._rpc_version_warning(17)
            kwargs["labels"] = list(labels)

        torrent_data = _try_read_torrent(torrent)

        if torrent_data:
            kwargs["metainfo"] = torrent_data
        else:
            kwargs["filename"] = torrent

        _rpc_version_check("torrent-add", kwargs, self.rpc_version)

        return list(self._request("torrent-add", kwargs, timeout=timeout).values())[0]

    def remove_torrent(self, ids: _TorrentIDs, delete_data: bool = False, timeout: _Timeout = None) -> None:
        """
        remove torrent(s) with provided id(s). Local data is removed if
        delete_data is True, otherwise not.
        """
        self._rpc_version_warning(3)
        self._request(
            "torrent-remove",
            {"delete-local-data": rpc_bool(delete_data)},
            ids,
            True,
            timeout=timeout,
        )

    def start_torrent(self, ids: _TorrentIDs, bypass_queue: bool = False, timeout: _Timeout = None) -> None:
        """Start torrent(s) with provided id(s)"""
        method = "torrent-start"
        if bypass_queue and self.rpc_version >= 14:
            method = "torrent-start-now"
        self._request(method, {}, ids, True, timeout=timeout)

    def start_all(self, bypass_queue: bool = False, timeout: _Timeout = None) -> None:
        """Start all torrents respecting the queue order"""
        torrent_list = self.get_torrents()
        method = "torrent-start"
        if self.rpc_version >= 14:
            if bypass_queue:
                method = "torrent-start-now"
            torrent_list = sorted(torrent_list, key=operator.attrgetter("queuePosition"))
        self._request(
            method,
            {},
            ids=[x.id for x in torrent_list],
            require_ids=True,
            timeout=timeout,
        )

    def stop_torrent(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """stop torrent(s) with provided id(s)"""
        self._request("torrent-stop", {}, ids, True, timeout=timeout)

    def verify_torrent(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """verify torrent(s) with provided id(s)"""
        self._request("torrent-verify", {}, ids, True, timeout=timeout)

    def reannounce_torrent(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """Reannounce torrent(s) with provided id(s)"""
        self._rpc_version_warning(5)
        self._request("torrent-reannounce", {}, ids, True, timeout=timeout)

    def get_torrent(
        self,
        torrent_id: _TorrentID,
        arguments: Iterable[str] = None,
        timeout: _Timeout = None,
    ) -> Torrent:
        """
        Get information for torrent with provided id.
        ``arguments`` contains a list of field names to be returned, when None
        all fields are requested. See the Torrent class for more information.

        new argument ``format`` in rpc_version 16 is unnecessarily
        and this lib can't handle table response, So it's unsupported.

        Returns a Torrent object with the requested fields.
        """
        if not arguments:
            arguments = self.torrent_get_arguments
        torrent_id = _parse_torrent_id(torrent_id)
        if torrent_id is None:
            raise ValueError("Invalid id")
        result: Dict[Union[str, int], Torrent] = self._request(
            "torrent-get",
            {"fields": arguments},
            torrent_id,
            require_ids=True,
            timeout=timeout,
        )
        if torrent_id in result:
            return result[torrent_id]
        for torrent in result.values():
            if torrent.hashString == torrent_id:
                return torrent
        raise KeyError("Torrent not found in result")

    def get_torrents(
        self,
        ids: _TorrentIDs = None,
        arguments: Iterable[str] = None,
        timeout: _Timeout = None,
    ) -> List[Torrent]:
        """
        Get information for torrents with provided ids. For more information see ``get_torrent``.

        Returns a list of Torrent object.
        """
        if arguments:
            arguments = list(set(arguments) | {"id"})
        else:
            arguments = self.torrent_get_arguments
        return list(self._request("torrent-get", {"fields": arguments}, ids, timeout=timeout).values())

    def get_files(
        self,
        ids: _TorrentIDs = None,
        timeout: _Timeout = None,
    ) -> Dict[int, List[File]]:
        """
        Get list of files for provided torrent id(s). If ids is empty,
        information for all torrents are fetched. This function returns a dictionary
        for each requested torrent id holding the information about the files.

        See more detail in :py:meth:`transmission_rpc.torrent.Torrent.files`

        .. code-block:: python

            {
                <torrent id>: [
                    <File 0>,
                    <File 1>,
                    ...
                ],
                ...
            }

        """
        fields = ["id", "name", "hashString", "files", "priorities", "wanted"]
        request_result: Dict[int, Torrent] = self._request("torrent-get", {"fields": fields}, ids, timeout=timeout)
        result = {}
        for tid, torrent in request_result.items():
            result[tid] = torrent.files()
        return result

    def set_files(self, items: Dict[str, Dict[str, Dict[str, Any]]], timeout: _Timeout = None) -> None:
        """
        Set file properties. Takes a dictionary with similar contents as the result
        of :py:meth:`transmission_rpc.client.Client.get_files`.

        .. code-block:: python

            {
                <torrent id>: {
                    <file id>: {
                        'priority': <priority ('high'|'normal'|'low')>,
                        'selected': <selected for download (True|False)>
                    },
                    ...
                },
                ...
            }

        """

        warn("deprecated", RemovedInV4Warning, stacklevel=2)
        if not isinstance(items, dict):
            raise ValueError("Invalid file description")
        for tid, files in items.items():
            if not isinstance(files, dict):
                continue
            wanted = []
            unwanted = []
            high = []
            normal = []
            low = []
            for fid, file_desc in files.items():
                if not isinstance(file_desc, dict):
                    continue
                if "selected" in file_desc and file_desc["selected"]:
                    wanted.append(fid)
                else:
                    unwanted.append(fid)
                if "priority" in file_desc:
                    if file_desc["priority"] == "high":
                        high.append(fid)
                    elif file_desc["priority"] == "normal":
                        normal.append(fid)
                    elif file_desc["priority"] == "low":
                        low.append(fid)

            self.change_torrent(
                [tid],
                timeout=timeout,
                priority_high=high or None,
                priority_normal=normal or None,
                priority_low=low or None,
                files_wanted=wanted or None,
                files_unwanted=unwanted or None,
            )

    def change_torrent(
        self,
        ids: _TorrentIDs,
        timeout: _Timeout = None,
        group: Optional[str] = None,
        tracker_list: Optional[Iterable[Iterable[str]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Change torrent parameters for the torrent(s) with the supplied id's. The
        parameters are:

        ============================ ===== =============== =============================================================
        Argument                     RPC   Replaced by     Description
        ============================ ===== =============== =============================================================
        ``bandwidthPriority``        5 -                   Priority for this transfer.
        ``downloadLimit``            5 -                   Set the speed limit for download in Kib/s.
        ``downloadLimited``          5 -                   Enable download speed limiter.
        ``files_unwanted``           1 -                   A list of file id's that shouldn't be downloaded.
        ``files_wanted``             1 -                   A list of file id's that should be downloaded.
        ``honorsSessionLimits``      5 -                   Enables or disables the transfer
                                                           to honour the upload limit set in the session.
        ``location``                 1 -                   Local download location.
        ``peer_limit``               1 -                   The peer limit for the torrents.
        ``priority_high``            1 -                   A list of file id's that should have high priority.
        ``priority_low``             1 -                   A list of file id's that should have normal priority.
        ``priority_normal``          1 -                   A list of file id's that should have low priority.
        ``queuePosition``            14 -                  Position of this transfer in its queue.
        ``seedIdleLimit``            10 -                  Seed inactivity limit in minutes.
        ``seedIdleMode``             10 -                  Seed inactivity mode. 0 = Use session limit,
                                                           1 = Use transfer limit, 2 = Disable limit.
        ``seedRatioLimit``           5 -                   Seeding ratio.
        ``seedRatioMode``            5 -                   Which ratio to use. 0 = Use session limit,
                                                           1 = Use transfer limit, 2 = Disable limit.
        ``speed_limit_down``         1 - 5 downloadLimit   Set the speed limit for download in Kib/s.
        ``speed_limit_down_enabled`` 1 - 5 downloadLimited Enable download speed limiter.
        ``speed_limit_up``           1 - 5 uploadLimit     Set the speed limit for upload in Kib/s.
        ``speed_limit_up_enabled``   1 - 5 uploadLimited   Enable upload speed limiter.
        ``trackerAdd``               10 -                  Array of string with announce URLs to add.
        ``trackerRemove``            10 -                  Array of ids of trackers to remove.
        ``trackerReplace``           10 -                  Array of (id, url) tuples
                                                           where the announce URL should be replaced.
        ``uploadLimit``              5 -                   Set the speed limit for upload in Kib/s.
        ``uploadLimited``            5 -                   Enable upload speed limiter.
        ``labels``                   16 -                  Array of string labels.
        ``group``                    17 -                  The name of this torrent's bandwidth group.
        ``tracker_list``             17 -                  A ``Iterable[Iterable[str]]``, each ``Iterable[str]`` for a tracker tier.
        ============================ ===== =============== =============================================================

        .. NOTE::

           transmission_rpc will try to automatically fix argument errors.

        """

        args = {}

        if tracker_list is not None:
            self._rpc_version_warning(17)
            args["trackerList"] = " ".join("\n".join(x) for x in tracker_list)

        if group is not None:
            self._rpc_version_warning(17)
            args["group"] = str(group)

        for key, value in kwargs.items():
            argument = make_rpc_name(key)
            arg, val = argument_value_convert("torrent-set", argument, value, self.rpc_version)
            args[arg] = val

        if len(args) > 0:
            self._request("torrent-set", args, ids, True, timeout=timeout)
        else:
            ValueError("No arguments to set")

    def move_torrent_data(
        self,
        ids: _TorrentIDs,
        location: Union[str, pathlib.Path],
        timeout: _Timeout = None,
    ) -> None:
        """Move torrent data to the new location."""
        self._rpc_version_warning(6)
        args = {"location": ensure_location_str(location), "move": True}
        self._request("torrent-set-location", args, ids, True, timeout=timeout)

    def locate_torrent_data(
        self,
        ids: _TorrentIDs,
        location: Union[str, pathlib.Path],
        timeout: _Timeout = None,
    ) -> None:
        """Locate torrent data at the provided location."""
        self._rpc_version_warning(6)
        args = {"location": ensure_location_str(location), "move": False}
        self._request("torrent-set-location", args, ids, True, timeout=timeout)

    def rename_torrent_path(
        self,
        torrent_id: _TorrentID,
        location: Union[str, pathlib.Path],
        name: str,
        timeout: _Timeout = None,
    ) -> Tuple[str, str]:
        """
        Rename directory and/or files for torrent.
        Remember to use get_torrent or get_torrents to update your file information.
        """
        self._rpc_version_warning(15)
        torrent_id = _parse_torrent_id(torrent_id)
        dirname = os.path.dirname(name)
        if len(dirname) > 0:
            raise ValueError("Target name cannot contain a path delimiter")
        args = {"path": ensure_location_str(location), "name": name}
        result = self._request("torrent-rename-path", args, torrent_id, True, timeout=timeout)
        return result["path"], result["name"]

    def queue_top(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """Move transfer to the top of the queue:_Timeout."""
        self._rpc_version_warning(14)
        self._request("queue-move-top", ids=ids, require_ids=True, timeout=timeout)

    def queue_bottom(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """Move transfer to the bottom of the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-bottom", ids=ids, require_ids=True, timeout=timeout)

    def queue_up(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """Move transfer up in the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-up", ids=ids, require_ids=True, timeout=timeout)

    def queue_down(self, ids: _TorrentIDs, timeout: _Timeout = None) -> None:
        """Move transfer down in the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-down", ids=ids, require_ids=True, timeout=timeout)

    def get_session(self, timeout: _Timeout = None) -> Session:
        """
        Get session parameters. See the Session class for more information.
        """
        self._request("session-get", timeout=timeout)
        self._update_server_version()
        return self.session

    def set_session(self, timeout: _Timeout = None, **kwargs: Any) -> None:
        """
        Set session parameters. The parameters are:

        ================================ ===== ================= ===========================================================
        Argument                         RPC   Replaced by       Description
        ================================ ===== ================= ===========================================================
        ``alt_speed_down``               5 -                     Alternate session download speed limit (in Kib/s).
        ``alt_speed_enabled``            5 -                     Enables alternate global download speed limiter.
        ``alt_speed_time_begin``         5 -                        Time when alternate speeds should be enabled.
                                                                    Minutes after midnight.
        ``alt_speed_time_day``           5 -                     Enables alternate speeds scheduling these days.
        ``alt_speed_time_enabled``       5 -                     Enables alternate speeds scheduling.
        ``alt_speed_time_end``           5 -                         Time when alternate speeds should be disabled.
                                                                     Minutes after midnight.
        ``alt_speed_up``                 5 -                     Alternate session upload speed limit (in Kib/s).
        ``blocklist_enabled``            5 -                     Enables the block list
        ``blocklist_url``                11 -                       Location of the block list.
                                                                    Updated with blocklist-update.
        ``cache_size_mb``                10 -                    The maximum size of the disk cache in MB
        ``dht_enabled``                  6 -                     Enables DHT.
        ``download_dir``                 1 -                     Set the session download directory.
        ``download_queue_enabled``       14 -                    Enables download queue.
        ``download_queue_size``          14 -                    Number of slots in the download queue.
        ``encryption``                   1 -                         Set the session encryption mode, one of ``required``,
                                                                     ``preferred`` or ``tolerated``.
        ``idle_seeding_limit``           10 -                    The default seed inactivity limit in minutes.
        ``idle_seeding_limit_enabled``   10 -                    Enables the default seed inactivity limit
        ``incomplete_dir``               7 -                     The path to the directory of incomplete transfer data.
        ``incomplete_dir_enabled``       7 -                         Enables the incomplete transfer data directory.

                                                                     Otherwise data for incomplete transfers are stored in
                                                                     the download target.
        ``lpd_enabled``                  9 -                     Enables local peer discovery for public torrents.
        ``peer_limit``                   1 - 5 peer-limit-global Maximum number of peers.
        ``peer_limit_global``            5 -                     Maximum number of peers.
        ``peer_limit_per_torrent``       5 -                     Maximum number of peers per transfer.
        ``peer_port``                    5 -                     Peer port.
        ``peer_port_random_on_start``    5 -                     Enables randomized peer port on start of Transmission.
        ``pex_allowed``                  1 - 5 pex-enabled       Allowing PEX in public torrents.
        ``pex_enabled``                  5 -                     Allowing PEX in public torrents.
        ``port``                         1 - 5 peer-port         Peer port.
        ``port_forwarding_enabled``      1 -                     Enables port forwarding.
        ``queue_stalled_enabled``        14 -                    Enable tracking of stalled transfers.
        ``queue_stalled_minutes``        14 -                    Number of minutes of idle that marks a transfer as stalled.
        ``rename_partial_files``         8 -                     Appends ".part" to incomplete files
        ``script_torrent_done_enabled``  9 -                     Whether or not to call the "done" script.
        ``script_torrent_done_filename`` 9 -                     Filename of the script to run when the transfer is done.
        ``seed_queue_enabled``           14 -                    Enables upload queue.
        ``seed_queue_size``              14 -                    Number of slots in the upload queue.
        ``seedRatioLimit``               5 -                     Seed ratio limit. 1.0 means 1:1 download and upload ratio.
        ``seedRatioLimited``             5 -                     Enables seed ration limit.
        ``speed_limit_down``             1 -                     Download speed limit (in Kib/s).
        ``speed_limit_down_enabled``     1 -                     Enables download speed limiting.
        ``speed_limit_up``               1 -                     Upload speed limit (in Kib/s).
        ``speed_limit_up_enabled``       1 -                     Enables upload speed limiting.
        ``start_added_torrents``         9 -                     Added torrents will be started right away.
        ``trash_original_torrent_files`` 9 -                     The .torrent file of added torrents will be deleted.
        ``utp_enabled``                  13 -                    Enables Micro Transport Protocol (UTP).
        ================================ ===== ================= ===========================================================

        .. NOTE::

           transmission_rpc will try to automatically fix argument errors.

        """
        args = {}
        for key, value in kwargs.items():
            if key == "encryption" and value not in [
                "required",
                "preferred",
                "tolerated",
            ]:
                raise ValueError("Invalid encryption value")
            argument = make_rpc_name(key)
            (arg, val) = argument_value_convert("session-set", argument, value, self.rpc_version)
            args[arg] = val
        if len(args) > 0:
            self._request("session-set", args, timeout=timeout)

    def blocklist_update(self, timeout: _Timeout = None) -> Optional[int]:
        """Update block list. Returns the size of the block list."""
        self._rpc_version_warning(5)
        result = self._request("blocklist-update", timeout=timeout)
        return result.get("blocklist-size")

    def port_test(self, timeout: _Timeout = None) -> Optional[bool]:
        """
        Tests to see if your incoming peer port is accessible from the
        outside world.
        """
        self._rpc_version_warning(5)
        result = self._request("port-test", timeout=timeout)
        return result.get("port-is-open")

    def free_space(self, path: Union[str, pathlib.Path], timeout: _Timeout = None) -> Optional[int]:
        """
        Get the amount of free space (in bytes) at the provided location.
        """
        self._rpc_version_warning(15)
        path = ensure_location_str(path)
        result: Dict[str, Any] = self._request("free-space", {"path": path}, timeout=timeout)
        if result["path"] == path:
            return result["size-bytes"]
        return None

    def session_stats(self, timeout: _Timeout = None) -> Session:
        """Get session statistics"""
        self._request("session-stats", timeout=timeout)
        return self.session

    def set_group(
        self,
        name: str,
        *,
        timeout: Optional[_Timeout] = None,
        honors_session_limits: Optional[bool] = None,
        speed_limit_down: Optional[int] = None,
        speed_limit_up_enabled: Optional[bool] = None,
        speed_limit_up: Optional[int] = None,
        speed_limit_down_enabled: Optional[bool] = None,
    ) -> None:
        self._rpc_version_warning(17)
        arguments: Dict[str, Any] = {"name": name}

        if honors_session_limits is not None:
            arguments["honorsSessionLimits"] = honors_session_limits

        if speed_limit_down is not None:
            arguments["speed-limit-down"] = speed_limit_down

        if speed_limit_up_enabled is not None:
            arguments["speed-limit-up-enabled"] = speed_limit_up_enabled

        if speed_limit_up is not None:
            arguments["speed-limit-up"] = speed_limit_up

        if speed_limit_down_enabled is not None:
            arguments["speed-limit-down-enabled"] = speed_limit_down_enabled

        self._request("group-get", arguments, timeout=timeout)

    def get_group(self, name: str, *, timeout: Optional[_Timeout] = None) -> Optional[Group]:
        self._rpc_version_warning(17)
        result: Dict[str, Any] = self._request("group-get", {"group": name}, timeout=timeout)

        if result["arguments"]["group"]:
            return Group(fields=result["arguments"]["group"][0])
        return None

    def get_groups(self, name: Optional[List[str]] = None, *, timeout: Optional[_Timeout] = None) -> Dict[str, Group]:
        payload = {}
        if name is not None:
            payload = {"group": name}

        result: Dict[str, Any] = self._request("group-set", payload, timeout=timeout)

        return {x["name"]: Group(fields=x) for x in result["arguments"]["group"]}

    def __enter__(self) -> "Client":
        return self

    def __exit__(
        self,
        exc_type: Type[Exception],
        exc_val: Exception,
        exc_tb: types.TracebackType,
    ) -> None:
        self._http_session.close()



# Copyright (c) 2020-2021 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
# pylint: disable=C0103
import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Union, Optional

from typing_extensions import TypedDict

from transmission_rpc.utils import format_timedelta
from transmission_rpc.constants import PRIORITY, IDLE_LIMIT, RATIO_LIMIT
from transmission_rpc.lib_types import File, Field, _Timeout

if TYPE_CHECKING:
    from transmission_rpc.client import Client


def get_status_old(code: int) -> str:
    """Get the torrent status using old status codes"""
    mapping = {
        (1 << 0): "check pending",
        (1 << 1): "checking",
        (1 << 2): "downloading",
        (1 << 3): "seeding",
        (1 << 4): "stopped",
    }
    return mapping[code]


_STATUS_NEW_MAPPING = {
    0: "stopped",
    1: "check pending",
    2: "checking",
    3: "download pending",
    4: "downloading",
    5: "seed pending",
    6: "seeding",
}


def get_status_new(code: int) -> str:
    """Get the torrent status using new status codes"""
    return _STATUS_NEW_MAPPING[code]


class Status(str):
    """A wrapped ``str`` for torrent status.

    returned by :py:attr:`.Torrent.status`
    """

    stopped: bool
    check_pending: bool
    checking: bool
    download_pending: bool
    downloading: bool
    seed_pending: bool
    seeding: bool

    def __new__(cls, raw: str) -> "Status":
        obj = super().__new__(cls, raw)
        for status in _STATUS_NEW_MAPPING.values():
            setattr(obj, status.replace(" ", "_"), raw == status)
        return obj


class Tracker(TypedDict):
    id: int
    announce: str
    scrape: str
    tier: int


class Torrent:
    """
    Torrent is a class holding the data received from Transmission regarding a bittorrent transfer.

    All fetched torrent fields are accessible through this class using attributes.
    This class has a few convenience properties using the torrent data.
    """

    def __init__(self, client: "Client", fields: Dict[str, Any]):
        if "id" not in fields:
            raise ValueError("Torrent requires an id")
        self._fields: Dict[str, Field] = {}
        self._update_fields(fields)
        self._incoming_pending = False
        self._outgoing_pending = False
        self._client = client

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """get the raw value from files by the **raw keys**"""
        try:
            return self._fields[key].value
        except KeyError:
            return default

    @property
    def id(self) -> int:
        """Returns the id for this torrent"""
        return self._fields["id"].value

    def _get_name_string(self) -> Optional[str]:
        """Get the name"""
        name = None
        # try to find name
        if "name" in self._fields:
            name = self._fields["name"].value
        return name

    def __repr__(self) -> str:
        tid = self._fields["id"].value
        name = self._get_name_string()
        if name is not None:
            return f'<Torrent {tid} "{name}">'
        return f"<Torrent {tid}>"

    def __str__(self) -> str:
        name = self._get_name_string()
        if name is not None:
            return f'Torrent "{name}"'
        return "Torrent"

    def __copy__(self) -> "Torrent":
        return Torrent(self._client, self._fields)

    def __getattr__(self, name: str) -> Any:
        try:
            return self._fields[name].value
        except KeyError:
            raise AttributeError(f"No attribute {name}") from None

    def _rpc_version(self) -> int:
        """Get the Transmission RPC API version."""
        if self._client:
            return self._client.rpc_version
        return 2

    def _dirty_fields(self) -> List[str]:
        """Enumerate changed fields"""
        outgoing_keys = [
            "bandwidthPriority",
            "downloadLimit",
            "downloadLimited",
            "peer_limit",
            "queuePosition",
            "seedIdleLimit",
            "seedIdleMode",
            "seedRatioLimit",
            "seedRatioMode",
            "uploadLimit",
            "uploadLimited",
        ]
        fields = []
        for key in outgoing_keys:
            if key in self._fields and self._fields[key].dirty:
                fields.append(key)
        return fields

    def _push(self) -> None:
        """Push changed fields to the server"""
        dirty = self._dirty_fields()
        args = {}
        for key in dirty:
            args[key] = self._fields[key].value
            self._fields[key] = self._fields[key]._replace(dirty=False)
        if len(args) > 0:
            self._client.change_torrent(self.id, **args)

    def _update_fields(self, other: Union["Torrent", Dict[str, Any]]) -> None:
        """
        Update the torrent data from a Transmission JSON-RPC arguments dictionary
        """
        if isinstance(other, dict):
            for key, value in other.items():
                self._fields[key.replace("-", "_")] = Field(value, False)
        elif isinstance(other, Torrent):
            for key in list(other._fields.keys()):
                self._fields[key] = Field(other._fields[key].value, False)
        else:
            raise ValueError("Cannot update with supplied data")
        self._incoming_pending = False

    def _status(self) -> str:
        """Get the torrent status"""
        code = self._fields["status"].value
        if self._rpc_version() >= 14:
            return get_status_new(code)
        return get_status_old(code)

    def files(self) -> List[File]:
        """
        Get list of files for this torrent.

        .. note ::


            The order of the files is guaranteed. The index of file object is the id of the file
            when calling :py:meth:`transmission_rpc.client.Client.set_files`.

        .. code-block:: python

            from transmission_rpc import Client

            torrent = Client().get_torrent(0)

            for file_id, file in enumerate(torrent.files()):
                print(file_id, file)

        """
        result: List[File] = []
        if "files" in self._fields:
            files = self._fields["files"].value
            indices = range(len(files))
            priorities = self._fields["priorities"].value
            wanted = self._fields["wanted"].value
            for item in zip(indices, files, priorities, wanted):
                selected = bool(item[3])
                priority = PRIORITY[item[2]]
                result.append(
                    File(
                        selected=selected,
                        priority=priority,
                        size=item[1]["length"],
                        name=item[1]["name"],
                        completed=item[1]["bytesCompleted"],
                    )
                )
        return result

    @property
    def name(self) -> str:
        """Returns the name of this torrent.

        Raise AttributeError if server don't return this field
        """
        return self.__getattr__("name")

    @property
    def status(self) -> Status:
        """
        :rtype: Status

        Returns the torrent status. Is either one of 'check pending', 'checking',
        'downloading', 'download pending', 'seeding', 'seed pending' or 'stopped'.
        The first two is related to verification.

        Examples:

        .. code-block:: python

            torrent = Torrent()
            torrent.status.downloading
            torrent.status == 'downloading'

        """
        return Status(self._status())

    @property
    def rateDownload(self) -> int:
        """
        Returns download rate in B/s

        :rtype: int
        """
        return self._fields["rateDownload"].value

    @property
    def rateUpload(self) -> int:
        """
        Returns upload rate in B/s

        :rtype: int
        """
        return self._fields["rateUpload"].value

    @property
    def hashString(self) -> str:
        """Returns the info hash of this torrent.

        :raise: AttributeError -- if server don't return this field
        """
        return self.__getattr__("hashString")

    @property
    def progress(self) -> float:
        """
        download progress in percent.

        :rtype: float
        """
        try:
            # https://gist.github.com/jackiekazil/6201722#gistcomment-2788556
            return round((100.0 * self._fields["percentDone"].value), 2)
        except KeyError:
            try:
                size = self._fields["sizeWhenDone"].value
                left = self._fields["leftUntilDone"].value
                return round((100.0 * (size - left) / float(size)), 2)
            except ZeroDivisionError:
                return 0.0

    @property
    def ratio(self) -> float:
        """
        upload/download ratio.

        :rtype: float
        """
        return float(self._fields["uploadRatio"].value)

    @property
    def eta(self) -> datetime.timedelta:
        """
        the "eta" as datetime.timedelta.

        If downloading, estimated the ``timedelta`` left until the torrent is done.
        If seeding, estimated the ``timedelta`` left until seed ratio is reached.

        raw `eta` maybe negative:
        - `-1` for ETA Not Available.
        - `-2` for ETA Unknown.

        https://github.com/transmission/transmission/blob/3.00/libtransmission/transmission.h#L1748-L1749

        :rtype: datetime.timedelta
        :raise ValueError: non positive ETA.
        """
        eta = self._fields["eta"].value
        if eta >= 0:
            return datetime.timedelta(seconds=eta)
        raise ValueError("eta not valid")

    @property
    def date_active(self) -> datetime.datetime:
        """the attribute ``activityDate`` as ``datetime.datetime`` in **UTC timezone**.

        .. note::

            raw ``activityDate`` value could be ``0`` for never activated torrent,
            therefore it can't always be converted to local timezone.


        :rtype: datetime.datetime
        """
        return datetime.datetime.fromtimestamp(self._fields["activityDate"].value, datetime.timezone.utc)

    @property
    def date_added(self) -> datetime.datetime:
        """raw field ``addedDate`` as ``datetime.datetime`` in **local timezone**.

        :rtype: datetime.datetime
        """
        return datetime.datetime.fromtimestamp(self._fields["addedDate"].value).astimezone()

    @property
    def date_started(self) -> datetime.datetime:
        """raw field ``startDate`` as ``datetime.datetime`` in **local timezone**.

        :rtype: datetime.datetime
        """
        return datetime.datetime.fromtimestamp(self._fields["startDate"].value).astimezone()

    @property
    def date_done(self) -> Optional[datetime.datetime]:
        """the attribute "doneDate" as datetime.datetime. returns None if "doneDate" is invalid."""
        done_date = self._fields["doneDate"].value
        # Transmission might forget to set doneDate which is initialized to zero,
        # so if doneDate is zero return None
        if done_date == 0:
            return None
        return datetime.datetime.fromtimestamp(done_date).astimezone()

    def format_eta(self) -> str:
        """
        Returns the attribute *eta* formatted as a string.

        * If eta is -1 the result is 'not available'
        * If eta is -2 the result is 'unknown'
        * Otherwise eta is formatted as <days> <hours>:<minutes>:<seconds>.
        """
        eta = self._fields["eta"].value
        if eta == -1:
            return "not available"
        if eta == -2:
            return "unknown"
        return format_timedelta(self.eta)

    @property
    def download_dir(self) -> Optional[str]:
        """The download directory.

        :available: transmission version 1.5.
        :available: RPC version 4.
        """
        return self._fields["downloadDir"].value

    @property
    def download_limit(self) -> Optional[int]:
        """The download limit.

        Can be a number or None.
        """
        if self._fields["downloadLimited"].value:
            return self._fields["downloadLimit"].value
        return None

    @download_limit.setter
    def download_limit(self, limit: int) -> None:
        """Download limit in Kbps or None."""
        if isinstance(limit, int):
            self._fields["downloadLimited"] = Field(True, True)
            self._fields["downloadLimit"] = Field(limit, True)
            self._push()
        elif limit is None:
            self._fields["downloadLimited"] = Field(False, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def peer_limit(self) -> int:
        """the peer limit."""
        return self._fields["peer_limit"].value

    @peer_limit.setter
    def peer_limit(self, limit: int) -> None:
        """
        Set the peer limit.
        """
        if isinstance(limit, int):
            self._fields["peer_limit"] = Field(limit, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def priority(self) -> str:
        """
        Bandwidth priority as string.
        Can be one of 'low', 'normal', 'high'. This is a mutator.
        """

        return PRIORITY[self._fields["bandwidthPriority"].value]

    @priority.setter
    def priority(self, priority: str) -> None:
        if isinstance(priority, str):
            self._fields["bandwidthPriority"] = Field(PRIORITY[priority], True)
            self._push()

    @property
    def seed_idle_limit(self) -> int:
        """
        seed idle limit in minutes.
        """
        return self._fields["seedIdleLimit"].value

    @seed_idle_limit.setter
    def seed_idle_limit(self, limit: int) -> None:
        """
        Set the seed idle limit in minutes.
        """
        if isinstance(limit, int):
            self._fields["seedIdleLimit"] = Field(limit, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def is_finished(self) -> bool:
        """Returns true if the torrent is finished (available from rpc version 2.0)"""
        return self._fields["isFinished"].value

    @property
    def is_stalled(self) -> bool:
        """Returns true if the torrent is stalled (available from rpc version 2.4)"""
        return self._fields["isStalled"].value

    @property
    def size_when_done(self) -> int:
        """Size in bytes when the torrent is done"""
        return self._fields["sizeWhenDone"].value

    @property
    def total_size(self) -> int:
        """Total size in bytes"""
        return self._fields["totalSize"].value

    @property
    def left_until_done(self) -> int:
        """Bytes left until done"""
        return self._fields["leftUntilDone"].value

    @property
    def desired_available(self) -> int:
        """Bytes that are left to download and available"""
        return self._fields["desiredAvailable"].value

    @property
    def error(self) -> int:
        """``0`` for fine task, non-zero for error torrent"""
        return self.__getattr__("error")

    @property
    def error_string(self) -> str:
        """empty string for fine task"""
        return self.__getattr__("errorString")

    @property
    def available(self) -> float:
        """Availability in percent"""
        bytes_all = self.total_size
        bytes_done = sum(x["bytesCompleted"] for x in self._fields["fileStats"].value)
        bytes_avail = self.desired_available + bytes_done
        return (bytes_avail / bytes_all) * 100 if bytes_all else 0

    @property
    def seed_idle_mode(self) -> str:
        """
        Seed idle mode as string. Can be one of 'global', 'single' or 'unlimited'.

         * global, use session seed idle limit.
         * single, use torrent seed idle limit. See seed_idle_limit.
         * unlimited, no seed idle limit.
        """
        return IDLE_LIMIT[self._fields["seedIdleMode"].value]

    @seed_idle_mode.setter
    def seed_idle_mode(self, mode: Union[str, int]) -> None:
        """
        Set the seed ratio mode.
        Can be one of ``global``, ``single`` or ``unlimited``, or ``0``, ``1``, ``2``.
        """
        if isinstance(mode, str):
            self._fields["seedIdleMode"] = Field(IDLE_LIMIT[mode], True)
            self._push()
        elif isinstance(mode, int):
            self._fields["seedIdleMode"] = Field(IDLE_LIMIT[mode], True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def seed_ratio_limit(self) -> float:
        """
        Torrent seed ratio limit as float. Also see seed_ratio_mode.
        This is a mutator.

        :rtype: float
        """

        return float(self._fields["seedRatioLimit"].value)

    @seed_ratio_limit.setter
    def seed_ratio_limit(self, limit: Union[int, float]) -> None:
        """
        Set the seed ratio limit as float.
        """
        if isinstance(limit, (int, float)) and limit >= 0.0:
            self._fields["seedRatioLimit"] = Field(float(limit), True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def seed_ratio_mode(self) -> str:
        """
        Seed ratio mode as string. Can be one of 'global', 'single' or 'unlimited'.

         * global, use session seed ratio limit.
         * single, use torrent seed ratio limit. See seed_ratio_limit.
         * unlimited, no seed ratio limit.

        This is a mutator.
        """
        return RATIO_LIMIT[self._fields["seedRatioMode"].value]

    @seed_ratio_mode.setter
    def seed_ratio_mode(self, mode: Union[str, int]) -> None:
        """
        Set the seed ratio mode.
        Can be one of ``'global'``, ``'single'`` or ``'unlimited'``, or ``0``, ``1``, ``2``.
        """
        if isinstance(mode, str):
            self._fields["seedRatioMode"] = Field(RATIO_LIMIT[mode], True)
            self._push()
        elif isinstance(mode, int):
            self._fields["seedRatioMode"] = Field(mode, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def upload_limit(self) -> Optional[int]:
        """
        upload limit.
        Can be a number or None.
        """
        if self._fields["uploadLimited"].value:
            return self._fields["uploadLimit"].value
        return None

    @upload_limit.setter
    def upload_limit(self, limit: Optional[int]) -> None:
        """Upload limit in Kbps or None."""
        if isinstance(limit, int):
            self._fields["uploadLimited"] = Field(True, True)
            self._fields["uploadLimit"] = Field(limit, True)
            self._push()
        elif limit is None:
            self._fields["uploadLimited"] = Field(False, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    @property
    def queue_position(self) -> int:
        """queue position for this torrent."""
        if self._rpc_version() >= 14:
            return self._fields["queuePosition"].value
        return 0

    @queue_position.setter
    def queue_position(self, position: str) -> None:
        """Queue position"""
        if self._rpc_version() >= 14:
            if isinstance(position, int):
                self._fields["queuePosition"] = Field(position, True)
                self._push()
            else:
                raise ValueError("Not a valid position")
        else:
            pass

    @property
    def trackers(self) -> List[Tracker]:
        return self.__getattr__("trackers")

    def update(self, timeout: _Timeout = None) -> None:
        """Update the torrent information."""
        self._push()
        torrent = self._client.get_torrent(self.id, timeout=timeout)
        self._update_fields(torrent)

    def start(self, bypass_queue: bool = False, timeout: _Timeout = None) -> None:
        """
        Start the torrent.
        """
        self._incoming_pending = True
        self._client.start_torrent(self.id, bypass_queue=bypass_queue, timeout=timeout)

    def stop(self, timeout: _Timeout = None) -> None:
        """Stop the torrent."""
        self._incoming_pending = True
        self._client.stop_torrent(self.id, timeout=timeout)

    def move_data(self, location: str, timeout: _Timeout = None) -> None:
        """Move torrent data to location."""
        self._incoming_pending = True
        self._client.move_torrent_data(self.id, location, timeout=timeout)

    def locate_data(self, location: str, timeout: _Timeout = None) -> None:
        """Locate torrent data at location."""
        self._incoming_pending = True
        self._client.locate_torrent_data(self.id, location, timeout=timeout)



# Copyright (c) 2018-2020 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
from transmission_rpc.error import TransmissionError
from transmission_rpc.client import Client
from transmission_rpc.session import Session
from transmission_rpc.torrent import Torrent
from transmission_rpc.constants import LOGGER, PRIORITY, RATIO_LIMIT, DEFAULT_TIMEOUT
from transmission_rpc.lib_types import File

__all__ = [
    "Client",
    "DEFAULT_TIMEOUT",
    "PRIORITY",
    "RATIO_LIMIT",
    "LOGGER",
    "TransmissionError",
    "Session",
    "Torrent",
    "File",
]



# Copyright (c) 2018-2021 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
from typing import TYPE_CHECKING, Any, Dict, Tuple, Union, Optional, Generator

from typing_extensions import Literal

from transmission_rpc.lib_types import Field

if TYPE_CHECKING:
    from transmission_rpc.client import Client


class Session:
    """
    Session is a dict-like class holding the session data for a Transmission daemon.

    Access the session field can be done through attributes.
    The attributes available are the same as the session arguments in the
    Transmission RPC specification, but with underscore instead of hyphen.


    get ``'download-dir'`` with ``session.download_dir``.

    .. code-block:: python

        session = Client().get_session()

        current = session.download_dir


    there are also setter like ``Session().download_dir = '/path/to/download'``

    .. code-block:: python

        session = Client().get_session()

        session.download_dir = '/path/to/new/download/dir'


    if you want to batch update a session, call ``.update(data)``

    .. code-block:: python

        session = Client().get_session()

        session.update({'k1': 'v1', "k2": "v2"})


    if you have to access to the private ``Session()._fields``,
    keys are stored with underscore.
    """

    def __init__(self, client: "Client", fields: Dict[str, Any] = None):
        self._client = client
        self._fields: Dict[str, Field] = {}
        if fields is not None:
            self._update(fields)

    def __getattr__(self, name: str) -> Any:
        try:
            return self._fields[name].value
        except KeyError as e:
            raise AttributeError(f"No attribute {name}") from e

    def _set(self, key: str, value: Any, commit: bool = False) -> None:
        key = key.replace("-", "_")
        current_field = self._fields.get(key)
        if current_field is None:
            self._fields[key] = Field(value, True)
        else:
            if current_field.value != value:
                self._fields[key] = Field(value, True)
        if commit:
            self._commit(key, value)

    def __str__(self) -> str:
        text = ""
        max_length = max(len(x) for x in self._fields.keys()) + 1
        for key, value in sorted(self._fields.items(), key=lambda x: x[0]):
            text += f"{key.ljust(max_length)}: {value.value!r}\n"
        return text

    def _commit(self, key: str = None, value: Any = None) -> None:
        """submit all dirty field to client"""
        dirty = {}

        if key is not None and value is not None:
            dirty[key] = value
        else:
            for k, v in self._fields.items():
                if v.dirty:
                    dirty[k] = v.value

        self._client.set_session(**dirty)

    def _update(self, other: Union[Dict[str, Any], "Session"]) -> None:
        if isinstance(other, dict):
            for key, value in other.items():
                self._set(key, value)
        elif isinstance(other, Session):
            for key, value in other._fields.items():
                self._set(key, value.value)
        else:
            raise ValueError("Cannot update with supplied data")

    def update(self, other: Union[Dict[str, Any], "Session"]) -> None:
        """
        Update the session data from a Transmission JSON-RPC arguments dictionary
        """
        self._update(other)
        self._commit()

    def keys(self) -> Generator[str, None, None]:
        """
        session keys with underscore (eg: ``download_dir``)
        """
        yield from self._fields.keys()

    def values(self) -> Generator[Any, None, None]:
        for value in self._fields.values():
            yield value.value

    def items(self) -> Generator[Tuple[str, Any], None, None]:
        """
        iter key,value pair

        hyphen in key is replace by underscore. (eg: ``'download_dir'``)
        """
        for key, field in self._fields.items():
            yield key, field.value

    @property
    def download_dir(self) -> str:
        """default download location

        - rpc version 12
        - transmission version 2.20
        :return:
        """
        return self.__getattr__("download_dir")

    @download_dir.setter
    def download_dir(self, location: str) -> None:
        """Enable/disable peer exchange."""
        if isinstance(location, str) and location:
            self._set("download_dir", location, True)
        else:
            raise TypeError(f"{location!r} if not a valid 'download-dir'")

    @property
    def version(self) -> str:
        """
        - rpc version 3
        - transmission version 1.41
        """
        return self.__getattr__("version")

    @property
    def rpc_version(self) -> int:
        """
        - rpc version 4
        - transmission version 1.50
        """
        return self.__getattr__("rpc_version")

    @property
    def peer_port(self) -> int:
        """Get the peer port.

        - rpc version 5
        - transmission version 1.60
        """
        return self.__getattr__("peer_port")

    @peer_port.setter
    def peer_port(self, port: int) -> None:
        """Set the peer port.

        - rpc version 5
        - transmission version 1.60
        """
        if isinstance(port, int):
            self._set("peer_port", port, True)
        else:
            raise ValueError("Not a valid limit")

    @property
    def pex_enabled(self) -> bool:
        """Is peer exchange enabled

        - rpc version 5
        - transmission version 1.60"""
        return self.__getattr__("pex_enabled")

    @pex_enabled.setter
    def pex_enabled(self, enabled: bool) -> None:
        """Enable/disable peer exchange."""
        if isinstance(enabled, bool):
            self._set("pex_enabled", enabled, True)
        else:
            raise TypeError("Not a valid type")

    @property
    def encryption(self) -> str:
        return self.__getattr__("encryption")

    @encryption.setter
    def encryption(self, value: Literal["required", "preferred", "tolerated"]) -> None:
        if value in {"required", "preferred", "tolerated"}:
            self._set("encryption", value, commit=True)
        else:
            raise ValueError("Not a valid encryption, can only be one of ['required', 'preferred', 'tolerated']")

    def get(self, key: str) -> Optional[Any]:
        v = self._fields.get(key)
        if v:
            return v.value
        return None

    @property
    def default_trackers(self) -> Optional[list]:
        """
        list of default trackers to use on public torrents
        new at rpc-version 17
        """
        return self.get("default-trackers")

    @property
    def rpc_version_semver(self) -> Optional[str]:
        """
        the current RPC API version in a semver-compatible str
        new at rpc-version 17
        """
        return self.get("rpc-version-semver")

    @property
    def script_torrent_added_enabled(self) -> Optional[bool]:
        """
        whether or not to call the `added` script
        new at rpc-version 17
        """
        return self.get("script-torrent-added-enabled")

    @property
    def script_torrent_added_filename(self) -> Optional[str]:
        """
        filename of the script to run
        new at rpc-version 17
        """
        return self.get("script-torrent-added-filename")

    @property
    def script_torrent_done_seeding_enabled(self) -> Optional[bool]:
        """
        whether or not to call the `seeding-done` script
        new at rpc-version 17
        """
        return self.get("script-torrent-done-seeding-enabled")

    @property
    def script_torrent_done_seeding_filename(self) -> Optional[str]:
        """
        filename of the script to run
        new at rpc-version 17
        """
        return self.get("script-torrent-done-seeding-filename")
